# Copyright 2001, 2002 SolidCode BV (http://www.solidcode.net)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package FontInfo;

use strict;
use pdfStream;
use Text::Iconv;
use Data::Dumper;

my %fonts=();		#fonts ass. array.
my @mediabox=();
my $rotate = 0;

sub getRotate {
	return $rotate;
}

sub getMediaBox {
	return (@mediabox);
}

sub LoadFontInfo{
	# LFI (pdffile, pageindex=0)
	 
	my $fname = shift(@_);
  	my ($pdf,@pages)=pdfStream::openpdf($fname);
	my $pagenr = ($_[0])?$_[0]:0;
	if ($pagenr > @pages) { die 'Pageindex $pagenr not found' };
	
	my $page = $pages[$pagenr];
	my $resources = $page->find_prop('Resources');

	my ($temp) = $page->find_prop('MediaBox')->val;
	@mediabox = ($temp->[0]->val,$temp->[1]->val,$temp->[2]->val,$temp->[3]->val);    
	$rotate = ($page->find_prop('Rotate'))?$page->find_prop('Rotate')->val:0;
	my $cc = Text::Iconv->new("UTF-16","UTF-8");
	
	foreach my $font (keys(%{$resources->{'Font'}->realise})) {
		if ($font =~ /^F\d+$|^R\d+$/) {
				 
			$resources->{'Font'}->{$font}->realise;
			($fonts{$font}{'BaseFont'}) = ($resources->{'Font'}->{$font}->{'BaseFont'}->val =~ /^(?:[A-Z]+\+)?(.*)$/);
			($fonts{$font}{'SubType'}) = $resources->{'Font'}->{$font}->{'Subtype'}->val;
			if ($resources->{'Font'}->{$font}->{'ToUnicode'}) { 
				($fonts{$font}{'Type'}) = $resources->{'Font'}->{$font}->{'ToUnicode'}->val =~ /^(?:[A-Z]+\+)?(.*)$/;
			}

			if ($resources->{'Font'}->{$font}->{'FirstChar'}) { 
				$fonts{$font}{'FirstChar'} = $resources->{'Font'}->{$font}->{'FirstChar'}->val;
				$fonts{$font}{'LastChar'} = $resources->{'Font'}->{$font}->{'LastChar'}->val;
				($fonts{$font}{'Widths'}) = $resources->{'Font'}->{$font}->{'Widths'}->val;
			}

			# == Embedded Fonts ==
			$fonts{$font}{'Embedded'} = (defined($resources->{'Font'}->{$font}->{'ToUnicode'})==1);
			if ($fonts{$font}{'Embedded'}) {
				# Read Unicode stream
				my $tounicode = $pdf->read_obj($resources->{'Font'}->{$font}->{'ToUnicode'});
				$tounicode->read_stream();
				# Create CMAP
				my ($cmap) = $tounicode->{' stream'} =~ /^.*beginbfrange\n(.*)endbfrange.*$/s;
				# split into rows 
				my (@cmap) = split(/\n/, "$cmap"); 
				my @keymap = [];
				# now map the ranges into final values
				for (my $i=0; $i<@cmap; $i++) {
					my @maprow = $cmap[$i] =~ /<([\dA-F]+)>/ig;
					for (my $k = hex($maprow[0]); $k <= hex($maprow[1]); $k++) {
						#
						my $val = hex($maprow[2]);
						$val += ($k - hex($maprow[0]));
						$keymap[$k] = $cc->convert(pack('v', $val));
						#
					}
				}
				@{$fonts{$font}{'CharCodes'}} = @keymap;
			}
		}
	}
	return %fonts;
}

sub dump {
	my $file = $_[0] || die 'Usage: Fontinfo::dump($pdffile, $pagenr)';
	my $page = ($_[1])?$_[1]:0;
	print "Fontinfo for $file at page $page\n";
	print '='x40,"\n";
	LoadFontInfo($file,$page);
}

1;
