# Copyright 2001, 2002 SolidCode BV (http://www.solidcode.net)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Gfxstate;

use strict;

my $debugging = 0;

sub debug {
	if ($debugging) { print STDERR "# ".$_[0]."\n"; } 
}

sub new {
    my $that  = shift;
    my $class = ref($that) || $that;
	my $self = {};
    bless($self, $class);
	$self->{LINEWIDTH}  = 1;
	$self->{DASHPATTERN}  = 0;
	$self->{PATHLIST} = [];
    return $self;
}

sub linewidth {
	my $self = shift;
	if (@_) { $self->{LINEWIDTH} = shift }
    return $self->{LINEWIDTH};
}

sub dashpattern {
	my $self = shift;
	if (@_) { $self->{DASHPATTERN} = shift }
    return $self->{DASHPATTERN};
}

sub startpoint {
	my $self = shift;
	if ( @{$self->{PATHLIST}} <= 0) { return undef; }
	else { return ${$self->{PATHLIST}}[0]->startpoint; }
}

sub endpoint {
	my $self = shift;
	if ( @{$self->{PATHLIST}} <= 0) { return undef; }
	else { return ${$self->{PATHLIST}}[0]->endpoint; }
}

sub isEmpty {
	my $self = shift;
	return (@{$self->{PATHLIST}} <= 0);
}

sub isClosed {
	my $self = shift;
	return ($self->startpoint == $self->endpoint);
}

sub push {
	debug("pushing");
	my $self = shift;
	push @{ $self->{PATHLIST} }, (shift)->deepcopy;
}

sub pop {
	debug("popping");
	my $self = shift;
	return pop @{$self->{PATHLIST}};
}

sub arshift {
	my $self = shift;
	return shift @{$self->{PATHLIST}};
}

sub clearpathlist {
	my $self = shift;
	$self->{PATHLIST} = [];
	return $self->{PATHLIST};
}

sub destroy {
	my $self = shift;
	$self->clearpathlist();
	$self->{LINEWIDTH} = undef;
	$self->{DASHPATTERN} = undef;
	$self = undef;
	return $self;
}

sub stroke {
	my $self = shift;
	my $color = shift;
	my @p = @{$self->{PATHLIST}};
	for (my $i=0; $i<@p; $i++) {
		$p[$i]->draw( 'none', $color, $self->linewidth, $self->dashpattern);
	}
}

sub fill {
	my $self = shift;
	my $color = shift;
	my @p = @{$self->{PATHLIST}};
	for (my $i=0; $i<@p; $i++) {
		$p[$i]->draw( $color, 'none', $self->linewidth);
	}
}

sub display {
	my $self = shift;
	print "Pathlist:\n".("="x20)."\n";
	print "Linewidth:\t".$self->{LINEWIDTH}."\n";
	print "DashPattern\t".$self->{DASHPATTERN}."\n";
	my @p = @{$self->{PATHLIST}};
	for (my $i=0; $i<@p; $i++) {
		print "PATH $i\n";
		$p[$i]->display;
	}
}

sub deepcopy {
	my $self  = shift;
	my $state = Gfxstate->new;
	while ((my $key, my $value) = each %{$self}) {
		$state->{$key} = $value;
	}
	return $state;
}
1;
