# Copyright 2001, 2002 SolidCode BV (http://www.solidcode.net)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Parser;

use strict;
use Math::Matrix;
use Text::Iconv;
use Scanner;
use pdfGraphics;
use FontInfo;

use Data::Dumper;

my $sym = ''; my @op = (); my $Ml;
my $Mr = new Math::Matrix([1, 0, 0], [0, 1, 0], [0, 0, 1]);
my $font_type = "";
my $font_type_max = "";
my $max_font_size_SF = 0;
my %fontlib;
my $text_lead = 0;
my $scaling_h = 100; #percentage
my $char_spacing = 0;
my $font_scaling = 0;
my $font_size = 0;
my $word_spacing = 0;
my $totalShift = 0;

# ---- text operators ----------------------------------------------------------#
sub extract {
	my $line = shift;
	my $Cmap = shift;
	my $newline = '';
	foreach my $chr (unpack("C*", $line)) {
			my $d = ($Cmap->[$chr])?$Cmap->[$chr]:"?";
			$newline .= $d;
	}
	return $newline;
}

# -----------------------------------------------------------------------------------

sub operator {
	$_ = $sym;
	if (/^Tj$/i) {
		#print "Tj << Mr: [$Mr->[0]->[0], $Mr->[0]->[1], $Mr->[0]->[2]], [$Mr->[1]->[0], $Mr->[1]->[1], $Mr->[1]->[2]], [$Mr->[2]->[0], $Mr->[2]->[1], $Mr->[2]->[2]]\n";
		#print "Tj << Ml: [$Ml->[0]->[0], $Ml->[0]->[1], $Ml->[0]->[2]], [$Ml->[1]->[0], $Ml->[1]->[1], $Ml->[1]->[2]], [$Ml->[2]->[0], $Ml->[2]->[1], $Ml->[2]->[2]]\n";
		
		$Mr = $Mr->multiply($Ml);  #!!!!
		my $boolRotate = ($Mr->[0]->[1] != 0 and $Mr->[1]->[0] != 0);
		my $extraStyle = (($boolRotate)?"rotate:on;":"");
		$extraStyle	  .= "letter-spacing: $char_spacing;";
		#pdfGraphics::printText($Mr->[2]->[0],$Mr->[2]->[1], $Mr->[0]->[0], 0, $extraStyle, $op[0]);
		pdfGraphics::printText($Mr->[2]->[0],$Mr->[2]->[1], $font_scaling * $font_size, 0, $extraStyle, $op[0]);

		#print "Tj >> Mr: [$Mr->[0]->[0], $Mr->[0]->[1], $Mr->[0]->[2]], [$Mr->[1]->[0], $Mr->[1]->[1], $Mr->[1]->[2]], [$Mr->[2]->[0], $Mr->[2]->[1], $Mr->[2]->[2]]\n";
		#print "Tj >> Ml: [$Ml->[0]->[0], $Ml->[0]->[1], $Ml->[0]->[2]], [$Ml->[1]->[0], $Ml->[1]->[1], $Ml->[1]->[2]], [$Ml->[2]->[0], $Ml->[2]->[1], $Ml->[2]->[2]]\n";
	} elsif (/^'|"$/) {
		#print "' << Mr: [$Mr->[0]->[0], $Mr->[0]->[1], $Mr->[0]->[2]], [$Mr->[1]->[0], $Mr->[1]->[1], $Mr->[1]->[2]], [$Mr->[2]->[0], $Mr->[2]->[1], $Mr->[2]->[2]]\n";
		#print "' << Ml: [$Ml->[0]->[0], $Ml->[0]->[1], $Ml->[0]->[2]], [$Ml->[1]->[0], $Ml->[1]->[1], $Ml->[1]->[2]], [$Ml->[2]->[0], $Ml->[2]->[1], $Ml->[2]->[2]]\n";
		
		$Ml->[2]->[0] = 0;
		$Ml->[2]->[1] = $text_lead * $Mr->[1]->[1];

		$Mr = $Mr->multiply($Ml);  #!!!!
		my $boolRotate = ($Mr->[0]->[1] != 0 and $Mr->[1]->[0] != 0);
		my $extraStyle = (($boolRotate)?"rotate:on;":"");
		$extraStyle	  .= "letter-spacing: $char_spacing;";
		pdfGraphics::printText($Mr->[2]->[0],$Mr->[2]->[1], $font_scaling * $font_size, 0, $extraStyle, $op[0]);

		#print "' >> Mr: [$Mr->[0]->[0], $Mr->[0]->[1], $Mr->[0]->[2]], [$Mr->[1]->[0], $Mr->[1]->[1], $Mr->[1]->[2]], [$Mr->[2]->[0], $Mr->[2]->[1], $Mr->[2]->[2]]\n";
		#print "' >> Ml: [$Ml->[0]->[0], $Ml->[0]->[1], $Ml->[0]->[2]], [$Ml->[1]->[0], $Ml->[1]->[1], $Ml->[1]->[2]], [$Ml->[2]->[0], $Ml->[2]->[1], $Ml->[2]->[2]]\n";
	} elsif (/^Td$/) {
		#print "Td << Mr: [$Mr->[0]->[0], $Mr->[0]->[1], $Mr->[0]->[2]], [$Mr->[1]->[0], $Mr->[1]->[1], $Mr->[1]->[2]], [$Mr->[2]->[0], $Mr->[2]->[1], $Mr->[2]->[2]]\n";
		#print "Td << Ml: [$Ml->[0]->[0], $Ml->[0]->[1], $Ml->[0]->[2]], [$Ml->[1]->[0], $Ml->[1]->[1], $Ml->[1]->[2]], [$Ml->[2]->[0], $Ml->[2]->[1], $Ml->[2]->[2]]\n";
		
		$Ml->[2]->[0] = $op[0] * $Mr->[0]->[0];
		$Ml->[2]->[1] = $op[1] * $Mr->[1]->[1];

		#print "Td >> Mr: [$Mr->[0]->[0], $Mr->[0]->[1], $Mr->[0]->[2]], [$Mr->[1]->[0], $Mr->[1]->[1], $Mr->[1]->[2]], [$Mr->[2]->[0], $Mr->[2]->[1], $Mr->[2]->[2]]\n";
		#print "Td >> Ml: [$Ml->[0]->[0], $Ml->[0]->[1], $Ml->[0]->[2]], [$Ml->[1]->[0], $Ml->[1]->[1], $Ml->[1]->[2]], [$Ml->[2]->[0], $Ml->[2]->[1], $Ml->[2]->[2]]\n";
	} elsif (/^TD$/) {
		$Ml->[2]->[0] = $op[0] * $Mr->[0]->[0];
		$Ml->[2]->[1] = $op[1] * $Mr->[1]->[1];
		$text_lead = $op[1];
	#} elsif (/^Tl$/) {
	} elsif (/^TL$/) {
		$text_lead = - $op[0];
	} elsif (/^T\*$/) {
		$Ml->[2]->[0] = 0;
		$Ml->[2]->[1] = $text_lead * $Mr->[1]->[1];
		#} elsif (/^['|"]$/) {
				#$Ml->[2]->[0] = 0;
				#$Ml->[2]->[1] = $text_lead * $Mr->[1]->[1];
				#$Mr = $Mr->multiply($Ml);
				#pdfGraphics::printText($Mr->[2]->[0..1], $Mr->[0]->[0], 0, '', $op[0]);
	} elsif (/^Tm$/) {
		$Ml = new Math::Matrix([1, 0, 0], [0, 1, 0], [0, 0, 1]);
		$Mr = new Math::Matrix([$op[0], $op[1], 0], [$op[2], $op[3], 0], [$op[4], $op[5], 1]);
		$font_size = $op[0] == 0 ? 1 : $op[0];

		if ($op[0] > $max_font_size_SF) {
			$max_font_size_SF = $op[0];
			$font_type_max = $font_type;
		}	
	} elsif (/^cm$/) {
		# translate & scale matrix
		$Ml = new Math::Matrix([1, 0, 0], [0, 1, 0], [0, 0, 1]);
		$Mr = new Math::Matrix([$op[0], $op[1], 0], [$op[2], $op[3], 0], [$op[4], $op[5], 1]);
	#} elsif (/^Trise$|^Ts$/) {
	} elsif (/^Th$/) {
		$scaling_h = $op[0];
	#} elsif (/^Tfs$/) {
			#$Ml->[0]->[0] = $op[0];
	} elsif (/^Tf$/) {
		#print "Tf << Mr: [$Mr->[0]->[0], $Mr->[0]->[1], $Mr->[0]->[2]], [$Mr->[1]->[0], $Mr->[1]->[1], $Mr->[1]->[2]], [$Mr->[2]->[0], $Mr->[2]->[1], $Mr->[2]->[2]]\n";
		#print "Tf << Ml: [$Ml->[0]->[0], $Ml->[0]->[1], $Ml->[0]->[2]], [$Ml->[1]->[0], $Ml->[1]->[1], $Ml->[1]->[2]], [$Ml->[2]->[0], $Ml->[2]->[1], $Ml->[2]->[2]]\n";
		
		$font_type = $op[0];
		pdfGraphics::setFontType($fontlib{$font_type}{"BaseFont"});
		#$Ml->[0]->[0] = $op[1];
		$font_scaling = $op[1];

		#print "Tf >> Mr: [$Mr->[0]->[0], $Mr->[0]->[1], $Mr->[0]->[2]], [$Mr->[1]->[0], $Mr->[1]->[1], $Mr->[1]->[2]], [$Mr->[2]->[0], $Mr->[2]->[1], $Mr->[2]->[2]]\n";
		#print "Tf >> Ml: [$Ml->[0]->[0], $Ml->[0]->[1], $Ml->[0]->[2]], [$Ml->[1]->[0], $Ml->[1]->[1], $Ml->[1]->[2]], [$Ml->[2]->[0], $Ml->[2]->[1], $Ml->[2]->[2]]\n";
	} elsif (/^Tc$/) {
		$char_spacing = $op[0];
	} elsif (/^Tw$/) {
		$word_spacing = $op[0];
	} elsif (/^BT$/) {
		$Ml = new Math::Matrix([1, 0, 0], [0, 1, 0], [0, 0, 1]);
	}
	#} elsif (/^ET$/) {


	# --- GFX ----
	elsif (/^Do$/) {
		# Do -> used for photographs and pictures
		# place filled rectangle for now
		my $breedte = $Mr->[0]->[0];
		my $hoogte = $Mr->[1]->[1];
		my $ypos = $Mr->[2]->[1]+$hoogte;
		pdfGraphics::setPoint($Mr->[2]->[0], $ypos);
		printf  "\t<rect class=\"photo\" x=\"%.2f\" y=\"%.2f\" width=\"%.2f\" height=\"%.2f\" style=\"fill:yellow; stroke: red\"\/>\n", $Mr->[2]->[0], Subpath::flipY($ypos), $breedte, $hoogte;
	}

	elsif (/^m$/) {
		# moveto 
		pdfGraphics::setPoint(@op);
	}
	elsif (/^l$/) {
		# lineto - append line-path
		pdfGraphics::addLine(@op);
	}
	elsif (/^n$/) {
		# end subpath without fill or stroke.
		# pdfGraphics::closePathListEmpty();
		pdfGraphics::clearPath();
	}
	elsif (/^h$/) {
		# close pathlist  with line.
		pdfGraphics::closePathList();
	}
	elsif (/^re$/) { 
		pdfGraphics::addRectangle(@op);
	}	
	elsif (/^c$/) {
		pdfGraphics::addCurve(@op);
	}	
	elsif (/^v$/) {
		pdfGraphics::addCurveB(@op);
	}
	elsif (/^y$/) {
		pdfGraphics::addCurveC(@op);
	}
	elsif (/^f*?$/i) { 
		# fill path (eo/nz)
		pdfGraphics::fillPathList();
		pdfGraphics::clearPathList();
	}	
	elsif (/^s$/) {
		# close and stroke path
		pdfGraphics::closePathList();
		pdfGraphics::strokePathList();
		pdfGraphics::clearPathList();
	} 
	elsif (/^S$/) {
		# stroke path
		pdfGraphics::strokePathList();
		pdfGraphics::clearPathList();
	} 
	elsif (/^B*?$/) {
		# fill & stroke path using evenodd/nonzero winding number rule
		pdfGraphics::fillPathList();
		pdfGraphics::strokePathList();
		pdfGraphics::clearPathList();
	} 
	elsif (/^b*?$/) {
		# close, fill & stroke path using evenodd/nonzero winding number rule
		pdfGraphics::closePathList();
		pdfGraphics::fillPathList();
		pdfGraphics::strokePathList();
		pdfGraphics::clearPathList();
	} 
	elsif (/^sh$/) {
		# Paint Area with shading pattern
		pdfGraphics::paintPathList();
		pdfGraphics::clearPathList();
	} 
	
	# graphic operators
	elsif (/^w$|^LW$/) {
		# set linewidth
		pdfGraphics::setLineWidth($op[0]);
	}
	elsif (/^d$/) {
		# set line dashpattern
		pdfGraphics::setDashPattern($op[0]);
	}
	elsif (/^q$/) { 
		# store gs
		pdfGraphics::storeState();
	}
	elsif (/^Q$/) { 
		# restore gs
		pdfGraphics::restoreState();
	}
	
	# Color operators
	elsif (/^gs$/) {
		# set graphics state parameters..
	}	
	elsif (/^cs$/) {
		# set non stroking colorspace
		pdfGraphics::setNonStrokingCS($op[0]);
	}
	elsif (/^CS$/) {
		# set stroking colorspace
		pdfGraphics::setStrokingCS($op[0]);
	}
	elsif (/^scn?$/) {
		pdfGraphics::setNonStrokingColor($op[0]);
	}
	elsif (/^SCN?$/) {
		pdfGraphics::setStrokingColor($op[0]);
	}
	elsif (/^g$/)  {
		pdfGraphics::setNonStrokingCS('DeviceGray');
		pdfGraphics::setNonStrokingColor($op[0]);
	}	
	elsif (/^G$/) {
		pdfGraphics::setStrokingCS('DeviceGray');
		pdfGraphics::setStrokingColor($op[0]);
	}	
	elsif (/^rg$/) {
		pdfGraphics::setNonStrokingCS('DeviceRGB');
		pdfGraphics::setNonStrokingColor(@op);
	}
	elsif (/^RG$/) {
		pdfGraphics::setStrokingCS('DeviceRGB');
		pdfGraphics::setStrokingColor(@op);
	}
	elsif (/^k$/) {
		pdfGraphics::setNonStrokingCS('DeviceCMYK');
		pdfGraphics::setNonStrokingColor($op[0]);
	}
	elsif (/^K$/) {
		pdfGraphics::setStrokingCS('DeviceCMYK');
		pdfGraphics::setStrokingColor(@op);
	}

	else {
			#print "-- syntax error: unknown operator: $_\n";
	}
}

sub parse {
	# Parse (stream, sourcefile, svgname, pagenr);
	# Init Scanner
	my $stream = shift(@_);
	Scanner::init($stream);
	
	# init font library, mediabox and rotating angle
	%fontlib = FontInfo::LoadFontInfo($_[0], $_[2]);
	my @mediaBox = FontInfo::getMediaBox();
	pdfGraphics::setMediaBox(@mediaBox);

	my $svgname = $_[1];
	my $viewBox = sprintf("%.2f %.2f %.2f %.2f", @mediaBox[0..3]);
	print "<?xml version=\"1.0\"?>\n";
	print "<svg width=\"".$mediaBox[2]."\" height=\"".$mediaBox[3]."\" style=\"font-family:Times New Roman\" viewBox=\"$viewBox\" xmlns:sm=\"http://www.solidcode.net/metadata\">\n";
    print "\t<sm:page name=\"$svgname\"/>\n";

	my $get = Scanner::stream_elements();
	my $c = Text::Iconv->new("MAC","UTF-8");
	
	while (defined($sym = &$get())) {
		if ($sym =~ /^[A-Za-z'"*]+$/) {
			operator(); undef(@op);
		} else {
			$totalShift = 0;
			#print STDERR "\"$sym\"\n";
			if ($sym =~ /^\s*\[|^\s*\(/) {
				if ($sym =~ /^\s*\[/) {
					my $s = '';
					foreach( $sym =~ /(\(.*?[^\\]\)|-?\d+(?:\.\d+)?)/g ) {
						my $treshold = 100;
						$s .= /^-?\d+(?:\.\d+)?$/ ? (($_ > $treshold || $_ < -$treshold) ? ' ' : '') : ($_ =~ s/^\((.*)\)$/$1/ ? $_ : $_);
						$s =~ s/\\\)/\)/g; $s =~ s/\\\(/\(/g;		#-- Replaces \( by ( and \) by )
						$totalShift += $_ if /^-?\d+(?:\.\d+)?$/
					}
					$sym = $s;
					#print STDERR "\t1:\"$sym\"\n";
				} elsif ($sym =~ /^\s*\(/) {
					$sym =~ s/\s*\((.*)\)\s*/$1/;
					$sym = $1;
					#print STDERR "\t2:\"$sym\"\n";
				}
				$sym =~ s/\\\)/\)/g;
				$sym =~ s/\\\(/\(/g;
				$sym =~ s/\\\]/\]/g;
				$sym =~ s/\\\[/\[/g;
				$sym =~ /[\x7f-\xff]/ && $sym =~ s/([\x7f-\xff])/$c->convert($1)/eg;
#				$sym =~ s/^\s+//; $sym =~ s/\s+$//;
				# convert embedded (mapped) characters
				if ($fontlib{$font_type}{'Embedded'}) {
					$sym =~ s/\\(\d{3})/chr($1-2)/eg;
					$sym = extract( $sym, $fontlib{$font_type}{'CharCodes'} );
				}
				else {
				# replacen van \xxx naar chr(\xxx)
					$sym =~ s/\\(\d{3})/"&#$1;"/eg; 
					#	$sym =~ s/\\(\d{3})/chr($1)/eg; 
					$sym =~ s/[\x00-\x19]/_/g;
				}
				# replace special html chars
				$sym =~ s/&/&amp;/g;
				$sym =~ s/</&lt;/g;
				$sym =~ s/>/&gt;/g;
			} elsif ($sym =~ /^\/(.*)$/) {
				#print STDERR "\t\"$_\"\n";
				$sym = $1;
			}
			push(@op, $sym);
		}
	}
	print "</svg>\n";
}

1;
