# Copyright 2001, 2002 SolidCode BV (http://www.solidcode.net)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Subpath;

use strict;
use Data::Dumper;

my $ymarge = 1600;

sub setYmarge {
	$ymarge = $_[0];
}

sub flipY {	
	# y-axe is reversed in svg
	return $ymarge-$_[0];
}

sub flipMultiY {
	# flip all y-axes in coordlist (so every odd index)
	my (@coords) = @_;
	for (my $i=1; $i<@coords; $i+=2) {
		$coords[$i] = flipY($coords[$i]);
	}
	return @coords;
}

sub new {
    my $that  = shift;
    my $class = ref($that) || $that;
	my $self = {};
    bless($self, $class);
	$self->{TYPE}  = undef;
	$self->{PATH} = [];
    return $self;
}
		
sub type {
	my $self = shift;
	return $self->{TYPE};
}

sub path {
	my $self = shift;
	return $self->{PATH};
}

sub setType {
	my $self = shift;
	my $type = shift;
	if ($type) { $self->{TYPE} = $type }
    return $self->{TYPE};
}

sub setPath {
	my $self = shift;

	$self->{PATH} = [];
	for (my $i=0; $i<@_; $i++) {
			$self->{PATH}[$i] = $_[$i];
	}
    return $self->{PATH};
}

sub set {
	my $self = shift;
	my $type = shift;
	my (@p) = @_;
	if ($type) { 
		$self->setType($type);
		$self->setPath(@p); 
	}
	return $self;
}

sub to_str {
	my $self = shift;
	my $str = "type:\t".(($self->{TYPE})?$self->{TYPE}:"undefined")."\n";
	$str .= "path:\t".((@{$self->{PATH}} <= 0)?"empty":"");
	for (my $i=0; $i<@{$self->{PATH}}; $i++) {
			$str .= "[".$self->{PATH}[$i]."] ";
	}
	$str .= "\n";
	return $str;
}

sub startpoint {
	my $self = shift;
	return @{$self->{PATH}}[0,1];
}

sub endpoint {
	my $self = shift;
	return @{$self->{PATH}}[-2,-1];
}

sub draw {
	# draw ( fillcolor, strokecolor, linewidth, dashpattern )
	my $self = shift;
	my $fillcolor   = ($_[0])?$_[0]:'none';
	my $strokecolor = ($_[1])?$_[1]:'none';
	my $linewidth   = ($_[2])?$_[2]:1;
	my $dashpattern = (!$_[3])?'':(($_[3]==0)?'solid':'dashed');
	my (@coords)    = @{$self->{PATH}};

	
	if ($self->type eq 'line') {
			printf "\t<line ";
			printf "x1=\"%.2f\" y1=\"%.2f\" x2=\"%.2f\" y2=\"%.2f\" ", flipMultiY(@coords[0..3]);
			printf "style=\"fill: %s; stroke: %s %s; stroke-width: %.2f\"",  $fillcolor, $dashpattern, $strokecolor, $linewidth;
			printf "/>\n";
	}
	elsif ($self->type eq 'rect') {
			$coords[1] = flipY($coords[1]);
		printf ("\t<rect x=\"%.2f\" y=\"%.2f\" width=\"%.2f\" height=\"%.2f\" style=\"fill: %s; stroke: %s; stroke-width: %.2f\"/>\n", $coords[0], $coords[1]-$coords[3], @coords[2..3], $fillcolor, $strokecolor, $linewidth);
	}
	elsif ($self->type eq 'curve') {
		@coords = flipMultiY(@coords);
		printf "\t<path d=\"M%.2f,%.2f C%.2f,%.2f %.2f,%.2f %.2f,%.2f\" ", @coords[0,1], @coords[0..5];
		printf "style=\"fill: %s; stroke: %s; stroke-width: %s;\"/>\n", $fillcolor, $strokecolor, $linewidth;
		#print "<!-- polygon style=\"fill: ".$fillcolor."; stroke: ".strokeColor.";\" points=\"";
	}
}

sub destroy {
	my $self = shift;
	$self->{PATH} = [];
}

sub deepcopy {
	my $self = shift;
	my $path = Subpath->new;
	while ((my $key, my $value) = each %{$self}) {
		$path->{$key} = $value;
	}
	return $path;
}

1;
