# Copyright 2002 Albert Reibenschuh - alfredreibenschuh@yahoo.com
# Adapted by SolidCode
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package pdfStream;

use strict;
use Text::PDF::File;
use Text::PDF::Utils;
use Data::Dumper;

sub proc_pages {
	my ($pdf, $pgs) = @_;
	my ($pg, $pgref, @pglist);
	my $pcount = 0;

	if(defined($pgs->{Resources})) {
		eval {
			$pgs->{Resources}->realise;
		};
	}
	foreach $pgref ($pgs->{'Kids'}->elementsof) {
		$pg = $pdf->read_obj($pgref);
		if ($pg->{'Type'}->val =~ m/^Pages$/o) {
			push(@pglist, proc_pages($pdf, $pg));
		} else {
			$pgref->{' pnum'} = $pcount++;
			if(defined($pg->{Resources})) {
				eval {
					$pg->{Resources}->realise;
				};
			}
			push (@pglist, $pgref);
		}
	}
	return(@pglist);
}

sub openpdf {
	my $file=shift(@_);
	my $pdf=Text::PDF::File->open($file);
	
	die("Couldn't open file: $file") unless $pdf;
	$pdf->{'Root'}->realise;
	$pdf->{'Root'}->{'Pages'}->realise;
	my @pages = proc_pages($pdf,$pdf->{'Root'}->{'Pages'});

	# van Erik, laten staan a.u.b.:     print Dumper(($pdf->{'Root'}->{'Pages'}->realise->{'Kids'}->elementsof)[0]->{'Resources'}->{'Font'}->realise);

	my @pagestack=sort {$a->{' pnum'} <=> $b->{' pnum'}} @pages;
	return($pdf,@pagestack);
}

sub fixcontents {
	my ($self) = @_;
        $self->{'Contents'} = $self->{'Contents'} || PDFArray();
        if(ref($self->{'Contents'})=~/Objind$/) {
		$self->{'Contents'}->realise;
	}
        if(ref($self->{'Contents'})!~/Array$/) {
	        $self->{'Contents'} = PDFArray($self->{'Contents'});
	}
}


sub streamdump { 			# dumps the stream of a pdf-page
	my $page=shift(@_);
	my $totalstream='';
	if(defined $page->{Contents}) {
		fixcontents($page);
		foreach my $k ($page->{Contents}->elementsof) {
			$k->realise if(ref($k)=~/Objind$/);

		 	my $stream=$k->{' stream'};

			if((defined $k->{Filter}) ) {

				# we need to fix filter because it MAY be
				# an array BUT IT COULD BE only a name
				if(ref($k->{Filter})!~/Array$/) {
				       $k->{Filter} = PDFArray($k->{Filter});
				}

				use Text::PDF::Filter;
				my @filts;
			        my ($hasflate) = -1;
			        my ($temp, $i, $temp1);
			        
			        for ($i = 0; $i <= $#{$k->{'Filter'}{' val'}}; $i++)
			        {
			            $temp = $k->{'Filter'}{' val'}[$i]->val;
			            $temp1 = "Text::PDF::$temp";
			            push (@filts, $temp1->new);
			        }

				foreach my $f (@filts) { 
						# print STDERR "Filter:  ",Dumper($f),"\n";
						$stream = $f->infilt($stream, 1); 
				}

			}
			$stream =~ s/\r/\n/g if $stream;
			$totalstream .= $stream if $stream;
		}
	}
	return $totalstream;
}

sub dump {
	my $fname = shift(@_);
  	my ($pdf,@pages)=openpdf($fname);
  	for (my $i=0; $i<@pages; $i++) {
			print "\n----- dumping page $i -----\n";
			print "\n".streamdump($pages[$i])."\n";
			print "\n".('-'x80)."\n";
	}
}

1;
